# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The catkin_tools build plugin is useful for building ROS parts.

This plugin relies on the catkin plugin as the catkin plugin performs
configuration that allows catkin_tools to run.  This plugin uses the
same keywords and configurations as the catkin plugin, the difference
is the installation of and using catkin_tools to build.
"""

import logging
import os

from snapcraft.plugins.v1 import catkin

logger = logging.getLogger(__name__)


class CatkinToolsPlugin(catkin.CatkinPlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["required"] = ["source"]
        return schema

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.stage_packages.append("python-catkin-tools")

        # Beta Warning
        # Remove this comment and warning once catkin tools plugin is stable.
        logger.warning(
            "The catkin tools plugin is currently in beta, "
            "its API may break. Use at your own risk"
        )

    def _prepare_build(self):
        super()._prepare_build()

        self._clean_catkin()

        self._add_catkin_profile()

        self._configure_catkin_profile()

    def _clean_catkin(self):
        # It's possible that this workspace wasn't initialized to be used with
        # catkin-tools, so initialize it first. Note that this is a noop if it
        # was already initialized.
        self.run(["catkin", "init"])

        # Clean to prevent conflicts with externally built packages.
        self.run(["catkin", "clean", "-y"])

    def _add_catkin_profile(self):
        # Overwrite the default catkin profile to ensure builds
        # aren't affected by profile changes.
        catkincmd = ["catkin", "profile", "add", "-f", "default"]

        self.run(catkincmd)

    def _configure_catkin_profile(self):
        # Use catkin config to set all configurations before running build.
        catkincmd = ["catkin", "config"]

        # Use the newly created default profile.
        catkincmd.extend(["--profile", "default"])

        # Configure catkin tools to use the snapcraft source, build, and
        # install directories.
        catkincmd.extend(["--build-space", self.builddir])

        catkincmd.extend(
            ["--source-space", os.path.join(self.builddir, self.options.source_space)]
        )

        catkincmd.extend(["--install", "--install-space", self.rosdir])

        # Add any cmake-args requested from the plugin options.
        catkincmd.append("--cmake-args")
        build_type = "Release"
        if "debug" in self.options.build_attributes:
            build_type = "Debug"
        catkincmd.append("-DCMAKE_BUILD_TYPE={}".format(build_type))

        catkincmd.extend(self._parse_cmake_args())

        self.run(catkincmd)

    def _build_catkin_packages(self):
        # Nothing to do if no packages were specified
        if self.catkin_packages is not None and len(self.catkin_packages) == 0:
            return

        # Call catkin build.
        catkincmd = ["catkin", "build"]

        # Prevent notification that the build is complete.
        catkincmd.append("--no-notify")

        # Use the newly created default profile.
        catkincmd.extend(["--profile", "default"])

        if self.catkin_packages:
            catkincmd.extend(self.catkin_packages)

        self.run(catkincmd)



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os
import shutil
import subprocess
import tempfile
import urllib.request


def main():
    try:
        temp_dir = tempfile.mkdtemp()
        compressed_snapcraft_source = download_snapcraft_source(temp_dir)
        compressed_snapcraft_sha256 = sha256_checksum(compressed_snapcraft_source)
        brew_formula_path = os.path.join(temp_dir, "snapcraft.rb")
        download_brew_formula(brew_formula_path)
        patched_dir = os.path.join(temp_dir, "patched")
        os.mkdir(patched_dir)
        brew_formula_from_source_path = os.path.join(patched_dir, "snapcraft.rb")
        patch_brew_formula_source(
            brew_formula_path,
            brew_formula_from_source_path,
            compressed_snapcraft_source,
            compressed_snapcraft_sha256,
        )
        install_brew_formula(brew_formula_from_source_path)
    finally:
        shutil.rmtree(temp_dir)


def download_snapcraft_source(dest_dir):
    dest_file = os.path.join(dest_dir, "snapcraft-0.1.tar.gz")
    branch_source = "https://github.com/{}/archive/{}.tar.gz".format(
        os.environ.get("TRAVIS_PULL_REQUEST_SLUG") or "snapcore/snapcraft",
        os.environ.get("TRAVIS_PULL_REQUEST_BRANCH") or "master",
    )
    print("Downloading branch source from {}".format(branch_source))
    urllib.request.urlretrieve(branch_source, dest_file)
    return dest_file


def sha256_checksum(filename):
    sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for block in iter(lambda: f.read(65536), b""):
            sha256.update(block)
    return sha256.hexdigest()


def download_brew_formula(destination_path):
    brew_formula_url = (
        "https://raw.githubusercontent.com/Homebrew/homebrew-core/master/"
        "Formula/snapcraft.rb"
    )
    urllib.request.urlretrieve(brew_formula_url, destination_path)


def patch_brew_formula_source(
    original_formula_path,
    destination_formula_path,
    compressed_source,
    compressed_sha256,
):
    with open(original_formula_path, "r") as original_file:
        with open(destination_formula_path, "w") as destination_file:
            for line in original_file:
                if line.startswith("  url "):
                    destination_file.write(
                        '  url "file://{}"\n'.format(compressed_source)
                    )
                elif line.startswith("  sha256 "):
                    destination_file.write('  sha256 "{}"\n'.format(compressed_sha256))
                else:
                    destination_file.write(line)


def install_brew_formula(formula_path):
    subprocess.check_call(["brew", "install", "--build-from-source", formula_path])


if __name__ == "__main__":
    main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016, 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import re
import urllib.parse
import uuid
from datetime import datetime
from typing import Any, Dict

import pymacaroons
from pyramid import response

from tests.fake_servers import base

logger = logging.getLogger(__name__)


class FakeStoreAPIServer(base.BaseFakeServer):

    _DEV_API_PATH = "/dev/api/"
    _V2_DEV_API_PATH = "/api/v2/snaps/"

    def __init__(self, fake_store, server_address):
        super().__init__(server_address)
        self.fake_store = fake_store
        self.account_keys = []
        self.registered_names = {}
        self.pushed_snaps = set()
        self.validation_set_revisions = dict()
        self.validation_set_assertions = dict()

    def configure(self, configurator):
        # POST
        configurator.add_route(
            "acl",
            urllib.parse.urljoin(self._DEV_API_PATH, "acl/"),
            request_method="POST",
        )
        configurator.add_view(self.acl, route_name="acl")

        configurator.add_route(
            "verify_acl",
            urllib.parse.urljoin(self._DEV_API_PATH, "acl/verify/"),
            request_method="POST",
        )
        configurator.add_view(self.verify_acl, route_name="verify_acl")

        configurator.add_route(
            "account_key",
            urllib.parse.urljoin(self._DEV_API_PATH, "account/account-key"),
            request_method="POST",
        )
        configurator.add_view(self.account_key, route_name="account_key")

        configurator.add_route(
            "snap_builds",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{dummy}/builds"),
            request_method="POST",
        )
        configurator.add_view(self.snap_builds, route_name="snap_builds")

        configurator.add_route(
            "snap_close",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{dummy}/close"),
            request_method="POST",
        )
        configurator.add_view(self.snap_close, route_name="snap_close")

        configurator.add_route(
            "snap_push",
            urllib.parse.urljoin(self._DEV_API_PATH, "snap-push/"),
            request_method="POST",
        )
        configurator.add_view(self.snap_push, route_name="snap_push")

        configurator.add_route(
            "snap_release",
            urllib.parse.urljoin(self._DEV_API_PATH, "snap-release/"),
            request_method="POST",
        )
        configurator.add_view(self.snap_release, route_name="snap_release")

        configurator.add_route(
            "register_name",
            urllib.parse.urljoin(self._DEV_API_PATH, "register-name/"),
            request_method="POST",
        )
        configurator.add_view(self.register_name, route_name="register_name")

        configurator.add_route(
            "agreement",
            urllib.parse.urljoin(self._DEV_API_PATH, "agreement/"),
            request_method="POST",
        )
        configurator.add_view(self.agreement, route_name="agreement")

        configurator.add_route(
            "snap_metadata_post",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/metadata"),
            request_method="POST",
        )
        configurator.add_view(self.snap_metadata, route_name="snap_metadata_post")

        configurator.add_route(
            "snap_binary_metadata_post",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/binary-metadata"),
            request_method="POST",
        )
        configurator.add_view(
            self.snap_binary_metadata, route_name="snap_binary_metadata_post"
        )

        configurator.add_route(
            "post_validation_sets_build_assertion",
            "/api/v2/validation-sets/build-assertion",
            request_method="POST",
        )
        configurator.add_view(
            self.post_validation_sets_build_assertion,
            route_name="post_validation_sets_build_assertion",
        )

        configurator.add_route(
            "post_validation_sets", "/api/v2/validation-sets", request_method="POST",
        )
        configurator.add_view(
            self.post_validation_sets, route_name="post_validation_sets",
        )

        # GET
        configurator.add_route(
            "get_validation_sets", "/api/v2/validation-sets", request_method="GET",
        )
        configurator.add_view(
            self.get_validation_sets, route_name="get_validation_sets",
        )

        configurator.add_route(
            "get_validation_sets_name",
            "/api/v2/validation-sets/{name}",
            request_method="GET",
        )
        configurator.add_view(
            self.get_validation_sets, route_name="get_validation_sets_name",
        )

        configurator.add_route(
            "details",
            urllib.parse.urljoin(self._DEV_API_PATH, "/details/upload-id/{snap}"),
            request_method="GET",
        )
        configurator.add_view(self.details, route_name="details")

        configurator.add_route(
            "account",
            urllib.parse.urljoin(self._DEV_API_PATH, "account"),
            request_method="GET",
        )
        configurator.add_view(self.account, route_name="account")

        configurator.add_route(
            "snap_state",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{dummy}/state"),
            request_method="GET",
        )
        configurator.add_view(self.snap_state, route_name="snap_state")

        configurator.add_route(
            "snap_channel_map",
            urllib.parse.urljoin(self._V2_DEV_API_PATH, "{snap}/channel-map"),
            request_method="GET",
        )
        configurator.add_view(self.snap_channel_map, route_name="snap_channel_map")

        configurator.add_route(
            "snap_releases",
            urllib.parse.urljoin(self._V2_DEV_API_PATH, "{snap}/releases"),
            request_method="GET",
        )
        configurator.add_view(self.snap_releases, route_name="snap_releases")

        configurator.add_route(
            "whoami", "/api/v2/tokens/whoami", request_method="GET",
        )

        configurator.add_view(self.whoami, route_name="whoami")

        configurator.add_route(
            "snap_validations",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/validations"),
            request_method="GET",
        )
        configurator.add_view(self.snap_validations, route_name="snap_validations")

        configurator.add_route(
            "snap_developers",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/developers"),
            request_method="GET",
        )
        configurator.add_view(self.snap_developers, route_name="snap_developers")

        configurator.add_route(
            "snap_binary_metadata_get",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/binary-metadata"),
            request_method="GET",
        )
        configurator.add_view(
            self.snap_binary_metadata, route_name="snap_binary_metadata_get"
        )

        # PUT
        configurator.add_route(
            "put_snap_validations",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/validations"),
            request_method="PUT",
        )
        configurator.add_view(
            self.put_snap_validations, route_name="put_snap_validations"
        )

        configurator.add_route(
            "put_snap_developers",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/developers"),
            request_method="PUT",
        )
        configurator.add_view(
            self.put_snap_developers, route_name="put_snap_developers"
        )

        configurator.add_route(
            "snap_metadata_put",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/metadata"),
            request_method="PUT",
        )
        configurator.add_view(self.snap_metadata, route_name="snap_metadata_put")

        configurator.add_route(
            "snap_binary_metadata_put",
            urllib.parse.urljoin(self._DEV_API_PATH, "snaps/{snap_id}/binary-metadata"),
            request_method="PUT",
        )
        configurator.add_view(
            self.snap_binary_metadata, route_name="snap_binary_metadata_put"
        )

    def _refresh_error(self):
        error = {
            "code": "macaroon-permission-required",
            "message": "Authorization Required",
        }
        payload = json.dumps({"error_list": [error]}).encode()
        response_code = 401
        content_type = "application/json"
        headers = [
            ("Content-Type", content_type),
            ("WWW-Authenticate", "Macaroon needs_refresh=1"),
        ]
        return response.Response(payload, response_code, headers)

    # POST

    def acl(self, request):
        content_type = "application/json"

        if "packages" in request.json_body:
            packages = request.json_body["packages"]
            unregistered_package = {"name": "unregistered-snap-name", "series": "16"}
            if unregistered_package in packages:
                return response.Response(
                    json.dumps(
                        {
                            "error_message": "Snap not found for the given snap "
                            "name: 'unregistered-snap-name' and "
                            "series: '16'"
                        }
                    ).encode(),
                    404,
                    [("Content-Type", content_type)],
                )

        permission = request.path.split("/")[-1]
        logger.debug("Handling ACL request for {}".format(permission))
        sso_host = urllib.parse.urlparse(
            os.environ.get("UBUNTU_ONE_SSO_URL", "http://localhost")
        ).netloc
        macaroon = pymacaroons.Macaroon(
            caveats=[
                pymacaroons.Caveat(
                    caveat_id="test caveat",
                    location=sso_host,
                    verification_key_id="test verifiacion",
                )
            ]
        )
        payload = json.dumps({"macaroon": macaroon.serialize()}).encode()
        response_code = 200
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def verify_acl(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()

        print(request.json_body)

        return self._verify_acl_wide_open()

    def _verify_acl_wide_open(self):
        acl = {
            "snap_ids": None,
            "channels": None,
            "permissions": ["package_upload", "package_access", "package_manage"],
        }

        payload = json.dumps(acl).encode()
        response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def account_key(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        data = request.json_body
        logger.debug("Handling account-key request with content {}".format(data))
        account_key_request = data["account_key_request"]

        if account_key_request == "test-not-implemented":
            return self._account_key_not_implemented()
        elif account_key_request == "test-invalid-data":
            return self._account_key_invalid_field()
        else:
            return self._account_key_successful(account_key_request)

    def _account_key_not_implemented(self):
        payload = b"Not Implemented"
        response_code = 501
        content_type = "text/plain"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _account_key_invalid_field(self):
        error = {
            "error_list": [
                {
                    "code": "invalid-field",
                    "message": "The account-key-request assertion is not valid.",
                }
            ]
        }
        payload = json.dumps(error).encode()
        response_code = 400
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _account_key_successful(self, account_key_request):
        # Extremely basic assertion parsing, just enough to make tests work.
        # Don't copy this.
        key_name = re.search(
            "^name: (.*)$", account_key_request, flags=re.MULTILINE
        ).group(1)
        key_id = re.search(
            "^public-key-sha3-384: (.*)$", account_key_request, flags=re.MULTILINE
        ).group(1)
        self.account_keys.append({"name": key_name, "public-key-sha3-384": key_id})
        account_key = {
            "account_key": {
                "account-id": "abcd",
                "name": key_name,
                "public-key-sha3-384": key_id,
            }
        }
        payload = json.dumps(account_key).encode()
        response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_builds(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling sign-build request")
        snap_build = request.json_body["assertion"]
        if snap_build == "test-not-implemented":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "feature-disabled",
                            "message": (
                                "The snap-build assertions are currently disabled."
                            ),
                        }
                    ]
                }
            ).encode()
            response_code = 501
            content_type = "application/json"
        elif snap_build == "test-invalid-data":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "invalid-field",
                            "message": "The snap-build assertion is not valid.",
                        }
                    ]
                }
            ).encode()
            response_code = 400
            content_type = "application/json"
        elif snap_build == "test-unexpected-data":
            payload = b"unexpected chunk of data"
            response_code = 500
            content_type = "text/plain"
        else:
            payload = json.dumps({"type": "snap-build", "foo": "bar"}).encode()
            response_code = 200
            content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_close(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling close request")
        channels = request.json_body["channels"]

        if channels == ["invalid"]:
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "invalid-field",
                            "message": "The 'channels' field content is not valid.",
                        }
                    ]
                }
            ).encode()
            response_code = 400
            content_type = "application/json"
        elif channels == ["unexpected"]:
            payload = b"unexpected chunk of data"
            response_code = 500
            content_type = "text/plain"
        elif channels == ["broken-plain"]:
            payload = b"plain data"
            response_code = 200
            content_type = "text/plain"
        elif channels == ["broken-json"]:
            payload = json.dumps({"closed_channels": channels}).encode()
            response_code = 200
            content_type = "application/json"
        else:
            payload = json.dumps(
                {
                    "closed_channels": channels,
                    "channel_map_tree": {
                        "latest": {
                            "16": {
                                "amd64": [
                                    {"channel": "stable", "info": "none"},
                                    {"channel": "candidate", "info": "none"},
                                    {
                                        "channel": "beta",
                                        "info": "specific",
                                        "version": "1.1",
                                        "revision": 42,
                                    },
                                    {"channel": "edge", "info": "tracking"},
                                ]
                            }
                        }
                    },
                }
            ).encode()
            response_code = 200
            content_type = "application/json"

        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_push(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug(
            "Handling upload request with content {}".format(request.json_body)
        )

        name = request.json_body["name"]
        if name == "test-snap-unregistered":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "resource-not-found",
                            "message": "the snap name does not exist",
                        }
                    ]
                }
            ).encode()
            response_code = 404
            content_type = "application/json"
        elif name == "test-snap-forbidden":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "resource-forbidden",
                            "message": "not allowed to publish this snap name",
                        }
                    ]
                }
            ).encode()
            response_code = 403
            content_type = "application/json"
        else:
            response_code = 202
            content_type = "application/json"
            if name == "test-review-snap":
                details_path = "details/upload-id/review-snap"
            elif name == "test-duplicate-snap":
                details_path = "details/upload-id/duplicate-snap"
            elif name == "test-scan-error-with-braces":
                details_path = "details/upload-id/scan-error-with-braces"
            else:
                details_path = "details/upload-id/good-snap"
            if not request.json_body.get("dry_run", False):
                snap_id = self.registered_names[name]["snap_id"]
                self.pushed_snaps.add(snap_id)
            payload = json.dumps(
                {
                    "status_details_url": urllib.parse.urljoin(
                        "http://localhost:{}/".format(self.server_port), details_path
                    )
                }
            ).encode()
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_release(self, request):  # noqa: C901
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug(
            "Handling release request with content {}".format(request.json_body)
        )
        response_code = 200
        content_type = "application/json"
        name = request.json_body["name"]
        channels = request.json_body["channels"]
        revision = request.json_body["revision"]
        if name == "test-snap-unregistered":
            response_code = 404
            content_type = "text/plain"
            payload = b""
        elif "alpha" in channels:
            response_code = 400
            payload = json.dumps({"errors": "Not a valid channel: alpha"}).encode()
        elif "edge/{curly}" in channels:
            response_code = 400
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "message": (
                                "Invalid branch name: {curly}. Enter a value consisting of letters, numbers or hyphens. "
                                "Hyphens cannot occur at the start or end of the chosen value."
                            ),
                            "code": "invalid-field",
                        }
                    ]
                }
            ).encode()
        elif "no-permission" in channels:
            response_code = 403
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "code": "macaroon-permission-required",
                            "message": "Permission is required: channel",
                        }
                    ],
                    "permission": "channel",
                    "channels": ["no-permission"],
                }
            ).encode()
        elif "bad-channel" in channels:
            response_code = 500
            payload = json.dumps({}).encode()
        elif name == "test-snap" or name.startswith("test-snapcraft"):
            payload = json.dumps(
                {
                    "opened_channels": channels,
                    "channel_map": [
                        {"channel": "stable", "info": "none"},
                        {"channel": "candidate", "info": "none"},
                        {
                            "revision": int(revision),
                            "channel": "beta",
                            "version": "0",
                            "info": "specific",
                        },
                        {"channel": "edge", "info": "tracking"},
                    ],
                }
            ).encode()
        elif name.startswith("arm-"):
            payload = json.dumps(
                {
                    "opened_channels": channels,
                    "channel_map_tree": {
                        "0.1": {
                            "16": {
                                "armhf": [
                                    {"channel": "stable", "info": "none"},
                                    {"channel": "candidate", "info": "none"},
                                    {
                                        "revision": int(revision),
                                        "channel": "beta",
                                        "version": "0",
                                        "info": "specific",
                                    },
                                    {"channel": "edge", "info": "tracking"},
                                ]
                            }
                        }
                    },
                }
            ).encode()
        elif name.startswith("multiarch-"):
            payload = json.dumps(
                {
                    "opened_channels": channels,
                    "channel_map_tree": {
                        "0.1": {
                            "16": {
                                "amd64": [
                                    {"channel": "stable", "info": "none"},
                                    {"channel": "candidate", "info": "none"},
                                    {
                                        "revision": int(revision),
                                        "channel": "beta",
                                        "version": "0",
                                        "info": "specific",
                                    },
                                    {"channel": "edge", "info": "tracking"},
                                ],
                                "armhf": [
                                    {"channel": "stable", "info": "none"},
                                    {"channel": "candidate", "info": "none"},
                                    {
                                        "revision": int(revision),
                                        "channel": "beta",
                                        "version": "0",
                                        "info": "specific",
                                    },
                                    {"channel": "edge", "info": "tracking"},
                                ],
                            }
                        }
                    },
                }
            ).encode()
        elif "notanumber" in revision:
            response_code = 400
            payload = json.dumps(
                {
                    "success": False,
                    "error_list": [
                        {
                            "code": "invalid-field",
                            "message": "The 'revision' field must be an integer",
                        }
                    ],
                    "errors": {"revision": ["This field must be an integer."]},
                }
            ).encode()
        else:
            raise NotImplementedError(
                "Cannot handle release request for {!r}".format(name)
            )
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def register_name(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug(
            "Handling registration request with content {}".format(request.json_body)
        )
        snap_name = request.json_body["snap_name"]
        store = request.json_body.get("store")

        if store == "snapcraft-test" and not snap_name.startswith(
            "test-snapcraft-in-store"
        ):
            return self._register_name_409_error("reserved_name")
        elif snap_name == "test-snap-name-already-registered":
            return self._register_name_409_error("already_registered")
        elif snap_name == "test-reserved-snap-name":
            return self._register_name_409_error("reserved_name")
        elif snap_name == "test-already-owned-snap-name":
            return self._register_name_409_error("already_owned")
        elif snap_name.startswith("test-snapcraft-fast"):
            return self._register_name_429_error("register_window")
        elif snap_name.startswith("test_invalid") or len(snap_name) > 40:
            return self._register_name_invalid(snap_name)
        elif snap_name == "snap-name-no-clear-error":
            return self._register_name_unclear_error()
        else:
            return self._register_name_successful(
                snap_name, request.json_body["is_private"]
            )

    def _register_name_409_error(self, error_code):
        payload = json.dumps(
            {
                "status": 409,
                "code": error_code,
                "register_name_url": "https://myapps.com/register-name/",
            }
        ).encode()
        response_code = 409
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _register_name_429_error(self, error_code):
        error = {"status": 429, "code": error_code}
        if error_code == "register_window":
            error["retry_after"] = 177
        payload = json.dumps(error).encode()
        response_code = 429
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _register_name_invalid(self, snap_name):
        # Emulates the current Store behaviour and never combines errors.
        if len(snap_name) > 40:
            msg = (
                "The name '{}' is not valid: it should be no longer than"
                " 40 characters."
            ).format(snap_name)
        elif snap_name.startswith("-") or snap_name.endswith("-"):
            msg = (
                "The name '{}' is not valid: it should not start"
                " nor end with a hyphen."
            ).format(snap_name)
        elif "--" in snap_name:
            msg = (
                "The name '{}' is not valid: it should not have"
                " two hyphens in a row."
            ).format(snap_name)
        else:
            msg = (
                "The name '{}' is not valid: it should only have"
                " ASCII lowercase letters, numbers, and hyphens,"
                " and must have at least one letter."
            ).format(snap_name)

        payload = json.dumps(
            {"error_list": [{"code": "invalid", "message": msg}]}
        ).encode()
        response_code = 400
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _register_name_unclear_error(self):
        payload = json.dumps({"status": 409, "code": "unexistent_error_code"}).encode()
        response_code = 409
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _register_name_successful(self, name, is_private):
        snap_id = uuid.uuid4().hex
        self.registered_names[name] = dict(private=is_private, snap_id=snap_id)
        payload = json.dumps({"snap_id": snap_id}).encode()
        response_code = 201
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def agreement(self, request):
        if "STORE_DOWN" in os.environ:
            response_code = 500
            content_type = "text/plain"
            payload = b"Broken"
        else:
            if request.json_body["latest_tos_accepted"] is not True:
                response_code = 400
                content_type = "application/json"
                payload = json.dumps(
                    {
                        "error_list": [
                            {
                                "message": "`latest_tos_accepted` must be `true`",
                                "code": "bad-request",
                                "extra": {"latest_tos_accepted": "true"},
                            }
                        ]
                    }
                ).encode()
            else:
                response_code = 200
                content_type = "application/json"
                payload = json.dumps(
                    {
                        "content": {
                            "latest_tos_accepted": True,
                            "tos_url": "http://fake-url.com",
                            "latest_tos_date": "2000-01-01",
                            "accepted_tos_date": "2010-10-10",
                        }
                    }
                ).encode()

        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_metadata(self, request):
        logger.debug("Handling metadata request")

        if "invalid" in request.json_body:
            err = {
                "error_list": [
                    {"message": "Invalid field: invalid", "code": "invalid-request"}
                ]
            }
            payload = json.dumps(err).encode("utf8")
            response_code = 400
            content_type = "application/json"
        elif any("conflict" in field_name for field_name in request.json_body):
            # conflicts!
            if request.method == "PUT":
                # update anyway
                payload = b""
                response_code = 200
                content_type = "text/plain"
            else:
                # POST, return error
                error_list = []
                for name, value in request.json_body.items():
                    if name == "test-conflict-with-braces":
                        conflict_value = "value with {braces}"
                    else:
                        conflict_value = value + "-changed"
                    error_list.append(
                        {
                            "message": "Conflict on {}".format(name),
                            "code": "conflict",
                            "extra": {
                                "name": name,
                                "field": name,
                                "current": conflict_value,
                                "value": value,
                            },
                        }
                    )
                payload = json.dumps({"error_list": error_list}).encode("utf8")
                response_code = 409
                content_type = "application/json"
        else:
            # all fine by default
            payload = b""
            response_code = 200
            content_type = "text/plain"

        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_binary_metadata(self, request):
        logger.debug("Handling binary metadata request")
        if request.method == "GET":
            current = [
                {"type": "icon", "hash": "1234567890", "filename": "icon.png"},
                {"type": "screenshot", "hash": "0987654321", "filename": "ss1.png"},
                {"type": "screenshot", "hash": "1122334455", "filename": "ss2.png"},
            ]
            return response.Response(
                json.dumps(current).encode("utf-8"),
                200,
                [("Content-Type", "application/json")],
            )
        else:
            # POST/PUT
            # snapcraft.storeapi._metadata._build_binary_request_data
            if type(request.params["info"]) == bytes:
                info = json.loads(request.params["info"].decode())
            else:
                info = json.loads(request.params["info"])
            invalid = any([e.get("filename", "").endswith("invalid") for e in info])
            conflict = any([e.get("filename", "").endswith("conflict") for e in info])
            conflict_with_braces = any(
                [e.get("filename", "").endswith("conflict-with-braces") for e in info]
            )
            if invalid:
                err = {
                    "error_list": [
                        {"message": "Invalid field: icon", "code": "invalid-request"}
                    ]
                }
                payload = json.dumps(err).encode("utf8")
                response_code = 400
            elif conflict and request.method == "POST":
                # POST, return error
                error_list = [
                    {
                        "message": "Conflict on icon",
                        "code": "conflict",
                        "extra": {
                            "name": "icon",
                            "field": "icon",
                            "current": "original-icon",
                            "value": "<posted-icon>",
                        },
                    }
                ]
                payload = json.dumps({"error_list": error_list}).encode("utf8")
                response_code = 409
            elif conflict_with_braces and request.method == "POST":
                # POST, return error
                error_list = [
                    {
                        "message": "Conflict on icon",
                        "code": "conflict",
                        "extra": {
                            "name": "icon",
                            "field": "icon",
                            "current": "original icon with {braces}",
                            "value": "<posted-icon>",
                        },
                    }
                ]
                payload = json.dumps({"error_list": error_list}).encode("utf8")
                response_code = 409
            else:
                updated_info = []
                for entry in info:
                    entry.pop("key", None)
                    updated_info.append(entry)
                payload = json.dumps(updated_info).encode("utf-8")
                response_code = 200

        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    # GET

    def details(self, request):
        snap = request.matchdict["snap"]
        if snap == "duplicate-snap":
            logger.debug("Handling duplicate snap request")
            payload = json.dumps(
                {
                    "code": "processing_error",
                    "url": "/dev/click-apps/5349/rev/1",
                    "can_release": False,
                    "revision": "1",
                    "processed": True,
                    "errors": [{"message": "Duplicate snap already uploaded"}],
                }
            ).encode()
        elif snap == "scan-error-with-braces":
            logger.debug("Handling request for scan error with braces")
            payload = json.dumps(
                {
                    "code": "processing_error",
                    "url": "/dev/click-apps/5349/rev/1",
                    "can_release": False,
                    "revision": "1",
                    "processed": True,
                    "errors": [{"message": "Error message with {braces}"}],
                }
            ).encode()
        else:
            logger.debug("Handling scan complete request")
            if snap == "good-snap":
                can_release = True
                code = "ready_to_release"
            elif snap == "review-snap":
                can_release = False
                code = "need_manual_review"
            payload = json.dumps(
                {
                    "code": code,
                    "url": "/dev/click-apps/5349/rev/1",
                    "can_release": can_release,
                    "revision": "1",
                    "processed": True,
                }
            ).encode()

        response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def account(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling account request")
        snaps = {
            "basic": {
                "snap-id": "snap-id",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "test-snap-with-no-validations": {
                "snap-id": "test-snap-id-with-no-validations",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "test-snap-with-dev": {
                "snap-id": "test-snap-id-with-dev",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "core": {
                "snap-id": "good",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "core-no-dev": {
                "snap-id": "no-dev",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "badrequest": {
                "snap-id": "badrequest",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "revoked": {
                "snap-id": "revoked",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "no-revoked": {
                "snap-id": "no-revoked",
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
            "no-id": {
                "snap-id": None,
                "status": "Approved",
                "private": False,
                "price": None,
                "since": "2016-12-12T01:01:01Z",
            },
        }
        snaps.update(
            {
                name: {
                    "snap-id": snap_data["snap_id"],
                    "status": "Approved",
                    "private": snap_data["private"],
                    "price": None,
                    "since": "2016-12-12T01:01:01Z",
                }
                for name, snap_data in self.registered_names.items()
            }
        )
        payload = json.dumps(
            {
                "account_id": "abcd",
                "account_keys": self.account_keys,
                "snaps": {"16": snaps},
            }
        ).encode()
        response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_channel_map(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling snap channel_map request")
        snap = request.matchdict["snap"]
        snap_channel_map = {
            "channel-map": [
                {
                    "architecture": "all",
                    "channel": "2.1/beta",
                    "expiration-date": None,
                    "revision": 1,
                    "progressive": {
                        "paused": None,
                        "percentage": None,
                        "current-percentage": None,
                    },
                    "when": "2020-02-03T20:58:37Z",
                }
            ],
            "revisions": [
                {
                    "architectures": [
                        "amd64",
                        "arm64",
                        "armhf",
                        "i386",
                        "s390x",
                        "ppc64el",
                    ],
                    "revision": 1,
                    "version": "10",
                }
            ],
            "snap": {
                "name": snap,
                "channels": [
                    {
                        "branch": None,
                        "fallback": None,
                        "name": "2.1/stable",
                        "risk": "stable",
                        "track": "2.1",
                    },
                    {
                        "branch": None,
                        "fallback": "2.1/stable",
                        "name": "2.1/candidate",
                        "risk": "candidate",
                        "track": "2.1",
                    },
                    {
                        "branch": None,
                        "fallback": "2.1/candidate",
                        "name": "2.1/beta",
                        "risk": "beta",
                        "track": "2.1",
                    },
                    {
                        "branch": None,
                        "fallback": "2.1/beta",
                        "name": "2.1/edge",
                        "risk": "edge",
                        "track": "2.1",
                    },
                ],
                "tracks": [
                    {
                        "name": "latest",
                        "status": "active",
                        "creation-date": None,
                        "version-pattern": None,
                    },
                    {
                        "name": "1.0",
                        "status": "default",
                        "creation-date": "2019-10-17T14:11:59Z",
                        "version-pattern": "1.*",
                    },
                ],
                "default-track": "2.1",
            },
        }

        response_code = 200
        content_type = "application/json"
        return response.Response(
            json.dumps(snap_channel_map).encode(),
            response_code,
            [("Content-Type", content_type)],
        )

    def snap_releases(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling snap releases request")
        snap_releases = {
            "revisions": [
                {
                    "architectures": ["arm64"],
                    "base": "core20",
                    "build_url": None,
                    "confinement": "strict",
                    "created_at": "2020-02-11T17:51:40.891996Z",
                    "grade": "stable",
                    "revision": 1,
                    "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
                    "size": 20,
                    "status": "Published",
                    "version": "1.0",
                },
            ],
            "releases": [
                {
                    "architecture": "amd64",
                    "branch": None,
                    "channel": "latest/edge",
                    "expiration-date": None,
                    "revision": 1,
                    "risk": "stable",
                    "track": "latest",
                    "when": "2020-01-12T17:51:40.891996Z",
                },
            ],
        }

        response_code = 200
        content_type = "application/json"
        return response.Response(
            json.dumps(snap_releases).encode(),
            response_code,
            [("Content-Type", content_type)],
        )

    def snap_state(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling snap state request")
        channel_map = {
            "channel_map_tree": {
                "latest": {
                    "16": {
                        "all": [
                            {"info": "none", "channel": "stable"},
                            {"info": "none", "channel": "candidate"},
                            {
                                "info": "specific",
                                "version": "1.1-amd64",
                                "channel": "beta",
                                "revision": 6,
                            },
                            {
                                "info": "specific",
                                "version": "1.0-i386",
                                "channel": "edge",
                                "revision": 3,
                            },
                            {
                                "info": "branch",
                                "version": "1.1-i386",
                                "channel": "edge/test",
                                "revision": 9,
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                        "i386": [
                            {"info": "none", "channel": "stable"},
                            {"info": "none", "channel": "candidate"},
                            {
                                "info": "specific",
                                "version": "1.1-amd64",
                                "channel": "beta",
                                "revision": 6,
                            },
                            {
                                "info": "specific",
                                "version": "1.0-i386",
                                "channel": "edge",
                                "revision": 3,
                            },
                            {
                                "info": "branch",
                                "version": "1.1-i386",
                                "channel": "edge/test",
                                "revision": 9,
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                        "amd64": [
                            {
                                "info": "specific",
                                "version": "1.0-amd64",
                                "channel": "stable",
                                "revision": 2,
                            },
                            {"info": "none", "channel": "candidate"},
                            {
                                "info": "specific",
                                "version": "1.1-amd64",
                                "channel": "beta",
                                "revision": 4,
                            },
                            {"info": "tracking", "channel": "edge"},
                            {
                                "info": "branch",
                                "version": "1.1-amd64",
                                "channel": "edge/test",
                                "revision": 10,
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                    }
                }
            }
        }

        parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(request.url).query)
        if "architecture" in parsed_qs:
            arch = parsed_qs["architecture"][0]
            series = channel_map["channel_map_tree"]["latest"]["16"]
            if arch in series:
                output = {"channel_map_tree": {"latest": {"16": {arch: series[arch]}}}}
            else:
                output = {}
        else:
            output = channel_map
        response_code = 200
        content_type = "application/json"
        return response.Response(
            json.dumps(output).encode(), response_code, [("Content-Type", content_type)]
        )

    def snap_validations(self, request):
        logger.debug("Handling validation request")
        snap_id = request.matchdict["snap_id"]
        if snap_id == "good":
            validation = [
                {
                    "approved-snap-id": "snap-id-1",
                    "approved-snap-revision": "3",
                    "approved-snap-name": "snap-1",
                    "authority-id": "dev-1",
                    "series": "16",
                    "sign-key-sha3-384": "1234567890",
                    "snap-id": "snap-id-gating",
                    "timestamp": "2016-09-19T21:07:27.756001Z",
                    "type": "validation",
                    "revoked": "false",
                    "required": True,
                },
                {
                    "approved-snap-id": "snap-id-2",
                    "approved-snap-revision": "5",
                    "approved-snap-name": "snap-2",
                    "authority-id": "dev-1",
                    "series": "16",
                    "sign-key-sha3-384": "1234567890",
                    "snap-id": "snap-id-gating",
                    "timestamp": "2016-09-19T21:07:27.756001Z",
                    "type": "validation",
                    "revoked": "false",
                    "required": False,
                },
                {
                    "approved-snap-id": "snap-id-3",
                    "approved-snap-revision": "-",
                    "approved-snap-name": "snap-3",
                    "authority-id": "dev-1",
                    "series": "16",
                    "sign-key-sha3-384": "1234567890",
                    "snap-id": "snap-id-gating",
                    "timestamp": "2016-09-19T21:07:27.756001Z",
                    "type": "validation",
                    "revoked": "false",
                    "required": True,
                },
            ]
            payload = json.dumps(validation).encode()
            response_code = 200
        elif snap_id == "bad":
            payload = "foo".encode()
            response_code = 200
        elif snap_id == "test-snap-id-with-no-validations":
            payload = json.dumps([]).encode()
            response_code = 200
        elif snap_id == "err":
            payload = json.dumps(
                {"error_list": [{"code": "test-code", "message": "test-error"}]}
            ).encode()
            response_code = 503
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def snap_developers(self, request):
        logger.debug("Handling snap developers request")
        snap_id = request.matchdict["snap_id"]
        if snap_id == "good":
            payload = json.dumps({"snap_developer": {}}).encode()
            response_code = 200
        elif snap_id in ("test-snap-id-with-dev", "revoked", "no-revoked"):
            payload = json.dumps(
                {
                    "snap_developer": {
                        "type": "snap-developer",
                        "authority-id": "dummy",
                        "publisher-id": "dummy",
                        "snap-id": snap_id,
                        "developers": [
                            {
                                "developer-id": "test-dev-id",
                                "since": "2017-02-10T08:35:00.390258Z",
                                "until": "2018-02-10T08:35:00.390258Z",
                            }
                        ],
                    }
                }
            ).encode()
            response_code = 200
        elif snap_id == "no-dev":
            payload = json.dumps(
                {
                    "error_list": [
                        {"message": "error", "code": "snap-developer-not-found"}
                    ]
                }
            ).encode()
            response_code = 403
        elif snap_id == "badrequest":
            payload = json.dumps({"snap_developer": {}}).encode()
            response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    # PUT

    def put_snap_validations(self, request):
        snap_id = request.matchdict["snap_id"]
        if snap_id == "good":
            payload = request.body
            response_code = 200
        elif snap_id == "err":
            payload = json.dumps(
                {"error_list": [{"code": "test-code", "message": "test-error"}]}
            ).encode()
            response_code = 501
        elif snap_id == "bad":
            payload = "foo".encode()
            response_code = 200
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def put_snap_developers(self, request):
        snap_id = request.matchdict["snap_id"]
        if snap_id in ("good", "test-snap-id-with-dev"):
            payload = request.body
            response_code = 200
        elif snap_id == "no-dev":
            payload = request.body
            response_code = 200
        elif snap_id == "badrequest":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "message": "The given `snap-id` does not match the "
                            "assertion.",
                            "code": "invalid-request",
                        }
                    ]
                }
            ).encode()
            response_code = 400
        elif snap_id == "revoked":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "message": "The assertion's `developers` would revoke "
                            "existing uploads.",
                            "code": "revoked-uploads",
                            "extra": ["this"],
                        }
                    ]
                }
            ).encode()
            response_code = 409
        elif snap_id == "no-revoked":
            payload = json.dumps(
                {
                    "error_list": [
                        {
                            "message": "The collaborators for this snap haven't been "
                            "altered. Exiting... ",
                            "code": "revoked-uploads",
                            "extra": ["this"],
                        }
                    ]
                }
            ).encode()
            response_code = 409
        content_type = "application/json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def whoami(self, request):
        if self.fake_store.needs_refresh:
            return self._refresh_error()
        logger.debug("Handling whoami request")
        payload = {
            "account": {
                "email": "foo@bar.baz",
                "id": "1234567890",
                "name": "Foo from Bar",
                "username": "baz",
            },
            "channels": None,
            "packages": None,
            "permissions": None,
        }

        response_code = 200
        content_type = "application/json"
        return response.Response(
            json.dumps(payload).encode(),
            response_code,
            [("Content-Type", content_type)],
        )

    def _get_validation_sets_revision(self, name: str) -> str:
        if name not in self.validation_set_revisions:
            self.validation_set_revisions[name] = 0
        else:
            self.validation_set_revisions[name] += 1

        return str(self.validation_set_revisions[name])

    def post_validation_sets_build_assertion(self, request):
        if request.json_body["sequence"] > 0:
            payload = request.json_body
            payload["sequence"] = str(payload["sequence"])
            # Get a revision for this assertion.
            payload["revision"] = self._get_validation_sets_revision(payload["name"])
            # Convert revisions for snaps into strings and add ids if not there.
            for snap in payload["snaps"]:
                if "revision" in snap:
                    snap["revision"] = str(snap["revision"])
                if "id" not in snap:
                    snap["id"] = "snap-id" + snap["name"]
            # Send and authority-id for this assertion.
            payload["authority-id"] = payload["account-id"]
            # And add all the other required fields for this response.
            payload["type"] = "validation-set"
            payload["timestamp"] = datetime.strftime(
                datetime.now(), "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            payload["series"] = "16"
            response_code = 200
        else:
            payload = {
                "error-list": [
                    {
                        "message": "0 is less than the minimum of 1 at /sequence",
                        "code": "invalid-request",
                    }
                ]
            }
            response_code = 400

        return response.Response(
            json.dumps(payload).encode(),
            response_code,
            [("Content-Type", "application/json")],
        )

    def _save_validation_sets_assertion(self, assertion: Dict[str, Any]) -> None:
        self.validation_set_assertions[assertion["name"]] = assertion

    def post_validation_sets(self, request):
        assertion_data = json.loads(request.body.decode().splitlines()[0])

        if assertion_data["authority-id"] == assertion_data["account-id"]:
            response_code = 200
            # API has a response with the assertion as json, but we don't use it.
            payload = {"assertions": [{"headers": assertion_data}]}
            self._save_validation_sets_assertion(assertion_data)
        else:
            payload = {
                "error-list": [
                    {
                        "message": "account-id and authority-id must match the requesting user.",
                        "code": "invalid-request",
                    }
                ]
            }
            response_code = 400

        return response.Response(
            json.dumps(payload).encode(),
            response_code,
            [("Content-Type", "application/json")],
        )

    def get_validation_sets(self, request):
        # This fake does not parse the sequence, it just verifies it is used
        # in some capacity.
        # TODO: support specific name.
        parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(request.url).query)

        try:
            name = request.matchdict["name"]
        except KeyError:
            name = None

        if name is not None and name in self.validation_set_assertions:
            assertions = [self.validation_set_assertions[name]]
        else:
            assertions = self.validation_set_assertions.values()

        if "sequence" not in parsed_qs or ["sequence"] == "latest":
            payload = {"assertions": [{"headers": a} for a in assertions]}
            response_code = 200
        else:
            payload = {
                "error-list": [
                    {
                        "message": "sequence must be one of ['all', 'latest'] or a positive integer higher than 0.",
                        "code": "invalid-request",
                    }
                ]
            }
            response_code = 400

        return response.Response(
            json.dumps(payload).encode(),
            response_code,
            [("Content-Type", "application/json")],
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This flutter plugin is useful for building flutter based parts.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - flutter-revision
      (string)
      Which Flutter revision to use for the build. This must be a valid
      revision from the flutter repository.
    - flutter-target
      (string, default: lib/main.dart)
      The main entry-point file of the application

This plugin works best with the flutter related extensions.
"""

import logging
from pathlib import Path
import subprocess
from typing import Any, Dict, List

from snapcraft import file_utils
from snapcraft.plugins.v1 import PluginV1

logger = logging.getLogger(__name__)


class FlutterPlugin(PluginV1):
    @classmethod
    def schema(cls) -> Dict[str, Any]:
        schema = super().schema()
        schema["properties"]["flutter-target"] = {
            "type": "string",
            "default": "lib/main.dart",
        }
        schema["properties"]["flutter-revision"] = {"type": "string", "default": None}
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls) -> List[str]:
        return ["flutter-revision"]

    @classmethod
    def get_build_properties(cls) -> List[str]:
        return ["flutter-target"]

    def __init__(self, name, options, project) -> None:
        super().__init__(name, options, project)

        logger.warning(
            "The flutter plugin is currently in beta, its API may break. Use at your "
            "own risk."
        )

    def pull(self) -> None:
        super().pull()

        work_path = Path(self.sourcedir)
        if self.options.source_subdir:
            work_path /= self.options.source_subdir

        # Let these errors go through to get them on Sentry.
        if self.options.flutter_revision:
            subprocess.run(
                f"yes | flutter version {self.options.flutter_revision}",
                shell=True,
                check=True,
                cwd=work_path,
            )

    def build(self) -> None:
        super().build()

        self.run(["flutter", "pub", "get"])

        self.run(
            [
                "flutter",
                "build",
                "linux",
                "--release",
                "-v",
                "-t",
                self.options.flutter_target,
            ]
        )

        # Flutter only supports arm64 and amd64
        if Path(self.builddir, "build/linux/x64/release/bundle").exists():
            bundle_dir_path = Path(self.builddir, "build/linux/x64/release/bundle")
        elif Path(self.builddir, "build/linux/arm64/release/bundle").exists():
            bundle_dir_path = Path(self.builddir, "build/linux/arm64/release/bundle")
        else:
            bundle_dir_path = Path(self.builddir, "build/linux/release/bundle")

        install_bin_dir_path = Path(self.installdir) / "bin"
        install_bin_dir_path.mkdir(exist_ok=True)
        # Now move everything over to the plugin's installdir
        file_utils.link_or_copy_tree(
            bundle_dir_path.as_posix(), install_bin_dir_path.as_posix()
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fixtures
from testtools.matchers import Equals

import snapcraft
from tests import unit


class VersionTestCase(unit.TestCase):
    def test_version_from_snap(self):
        self.useFixture(fixtures.EnvironmentVariable("SNAP_NAME", "snapcraft"))
        self.useFixture(fixtures.EnvironmentVariable("SNAP_VERSION", "3.14"))
        self.assertThat(snapcraft._get_version(), Equals("3.14"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Equals

from snapcraft import project
from snapcraft.internal.project_loader import inspection
from tests import unit


class ProvidesTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.part1 = self.load_part("part1")
        self.part2 = self.load_part("part2")
        self.parts = [self.part1, self.part2]
        self.project = project.Project()

        self.part1.makedirs()
        self.part2.makedirs()

    def test_provides_stage(self):
        file1_path = os.path.join(self.project.stage_dir, "file1")
        file2_path = os.path.join(self.project.stage_dir, "file2")

        self.part1.mark_stage_done({"file1"}, set())
        self.part2.mark_stage_done({"file2"}, set())

        for file_path in (file1_path, file2_path):
            open(file_path, "w").close()

        self.assertThat(
            inspection.provides(file1_path, self.project, self.parts),
            Equals({self.part1}),
        )
        self.assertThat(
            inspection.provides(file2_path, self.project, self.parts),
            Equals({self.part2}),
        )

    def test_provides_prime(self):
        file1_path = os.path.join(self.project.prime_dir, "file1")
        file2_path = os.path.join(self.project.prime_dir, "file2")

        self.part1.mark_prime_done({"file1"}, set(), set(), set())
        self.part2.mark_prime_done({"file2"}, set(), set(), set())

        open(file1_path, "w").close()
        open(file2_path, "w").close()

        self.assertThat(
            inspection.provides(file1_path, self.project, self.parts),
            Equals({self.part1}),
        )
        self.assertThat(
            inspection.provides(file2_path, self.project, self.parts),
            Equals({self.part2}),
        )

    def test_provides_dir(self):
        dir_path = os.path.join(self.project.stage_dir, "dir")
        file1_path = os.path.join(dir_path, "file1")
        file2_path = os.path.join(dir_path, "file2")

        self.part1.mark_stage_done({os.path.join("dir", "file1")}, {"dir"})
        self.part2.mark_stage_done({os.path.join("dir", "file2")}, {"dir"})

        for file_path in (file1_path, file2_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            open(file_path, "w").close()

        self.assertThat(
            inspection.provides(dir_path, self.project, self.parts),
            Equals({self.part1, self.part2}),
        )

    def test_provides_file_outside_stage_or_prime(self):
        file_path = os.path.join(self.part1.plugin.sourcedir, "file")
        open(file_path, "w").close()

        raised = self.assertRaises(
            inspection.errors.ProvidesInvalidFilePathError,
            inspection.provides,
            file_path,
            self.project,
            self.parts,
        )

        self.assertThat(raised.path, Equals(file_path))

    def test_provides_untracked_file(self):
        file_path = os.path.join(self.project.stage_dir, "file")
        open(file_path, "w").close()

        raised = self.assertRaises(
            inspection.errors.UntrackedFileError,
            inspection.provides,
            file_path,
            self.project,
            self.parts,
        )

        self.assertThat(raised.path, Equals(file_path))

    def test_provides_no_such_file(self):
        file_path = os.path.join(self.project.stage_dir, "foo")

        raised = self.assertRaises(
            inspection.errors.NoSuchFileError,
            inspection.provides,
            file_path,
            self.project,
            self.parts,
        )

        self.assertThat(raised.path, Equals(file_path))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.plugins.v2.dump import DumpPlugin


class DumpPluginTest(TestCase):
    def test_schema(self):
        schema = DumpPlugin.get_schema()

        self.assertThat(
            schema["$schema"], Equals("http://json-schema.org/draft-04/schema#")
        )
        self.assertThat(schema["type"], Equals("object"))
        self.assertFalse(schema["additionalProperties"])
        self.assertFalse(schema["properties"])

    def test_get_build_packages(self):
        plugin = DumpPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_packages(), Equals(set()))

    def test_get_build_environment(self):
        plugin = DumpPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_environment(), Equals(dict()))

    def test_get_build_commands(self):
        plugin = DumpPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                ['cp --archive --link --no-dereference . "${SNAPCRAFT_PART_INSTALL}"']
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
import tarfile
import tempfile

import debian.arfile

from . import errors
from ._base import FileBase


class Deb(FileBase):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
        )
        if source_tag:
            raise errors.SnapcraftSourceInvalidOptionError("deb", "source-tag")
        elif source_commit:
            raise errors.SnapcraftSourceInvalidOptionError("deb", "source-commit")
        elif source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("deb", "source-branch")

    def provision(self, dst, clean_target=True, keep_deb=False, src=None):
        if src:
            deb_file = src
        else:
            deb_file = os.path.join(self.source_dir, os.path.basename(self.source))

        if clean_target:
            tmp_deb = tempfile.NamedTemporaryFile().name
            shutil.move(deb_file, tmp_deb)
            shutil.rmtree(dst)
            os.makedirs(dst)
            shutil.move(tmp_deb, deb_file)

        # Importing DebFile causes LP: #1731478 when snapcraft is
        # run as a snap.
        deb_ar = debian.arfile.ArFile(deb_file)
        try:
            data_member_name = [
                i for i in deb_ar.getnames() if i.startswith("data.tar")
            ][0]
        except IndexError:
            raise errors.InvalidDebError(deb_file=deb_file)
        data_member = deb_ar.getmember(data_member_name)
        with tarfile.open(fileobj=data_member) as tar:
            tar.extractall(dst)

        if not keep_deb:
            os.remove(deb_file)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform

import pytest

from snapcraft.plugins.v2.conda import (
    CondaPlugin,
    ArchitectureMissing,
    _get_miniconda_source,
)


def test_schema():
    assert CondaPlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "conda-packages": {
                "type": "array",
                "minItems": 1,
                "uniqueItems": True,
                "items": {"type": "string"},
                "default": [],
            },
            "conda-python-version": {"type": "string", "default": ""},
            "conda-miniconda-version": {"type": "string", "default": "latest"},
        },
    }


def test_get_build_packages():
    plugin = CondaPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_packages() == {"curl"}


def test_get_build_environment():
    plugin = CondaPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_environment() == {"PATH": "${HOME}/miniconda/bin:${PATH}"}


def test_get_build_commands():
    class Options:
        conda_packages = ["python"]
        conda_python_version = ""
        conda_miniconda_version = "latest"

    plugin = CondaPlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'if ! [ -e "${HOME}/miniconda.sh" ]; then\n    curl --proto \'=https\' --tlsv1.2 -sSf https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh > ${HOME}/miniconda.sh\n    chmod 755 ${HOME}/miniconda.sh\n    export PATH="${HOME}/miniconda/bin:${PATH}"\nfi\n',
        "${HOME}/miniconda.sh -bfp ${HOME}/miniconda",
        "CONDA_TARGET_PREFIX_OVERRIDE=/snap/${SNAPCRAFT_PROJECT_NAME}/current conda create --prefix $SNAPCRAFT_PART_INSTALL --yes python",
    ]


def test_get_build_commands_with_python_version_parameters():
    class Options:
        conda_packages = ["python"]
        conda_python_version = "3.7"
        conda_miniconda_version = "latest"

    plugin = CondaPlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'if ! [ -e "${HOME}/miniconda.sh" ]; then\n    curl --proto \'=https\' --tlsv1.2 -sSf https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh > ${HOME}/miniconda.sh\n    chmod 755 ${HOME}/miniconda.sh\n    export PATH="${HOME}/miniconda/bin:${PATH}"\nfi\n',
        "${HOME}/miniconda.sh -bfp ${HOME}/miniconda",
        "CONDA_TARGET_PREFIX_OVERRIDE=/snap/${SNAPCRAFT_PROJECT_NAME}/current conda create --prefix $SNAPCRAFT_PART_INSTALL --yes python=3.7 python",
    ]


@pytest.mark.parametrize(
    "arch",
    [("amd64", "x86_64"), ("ppc64el", "ppc64le"), ("armhf", "armv7l"), ("i386", "x86")],
)
def test_get_miniconda_source_with_snap_arch(monkeypatch, arch):
    monkeypatch.setenv("SNAP_ARCH", arch[0])

    assert (
        _get_miniconda_source(version="1.9")
        == f"https://repo.anaconda.com/miniconda/Miniconda3-1.9-Linux-{arch[1]}.sh"
    )


def test_get_miniconda_source_unsupported_snap_arch(monkeypatch):
    # raising False as this may only be set by a running
    # environment triggering the tests.
    monkeypatch.setenv("SNAP_ARCH", "arm64")

    with pytest.raises(ArchitectureMissing):
        _get_miniconda_source(version="1.9")


def test_get_miniconda_source_unsupported_dev_env(monkeypatch):
    monkeypatch.setattr(platform, "machine", lambda: "foo")
    monkeypatch.delenv("SNAP_ARCH", raising=False)

    with pytest.raises(KeyError):
        _get_miniconda_source(version="1.9")


def test_architecture_missing_exception():
    exception = ArchitectureMissing("amd64")

    assert (
        exception.get_brief()
        == "Architecture 'amd64' is not supported with the 'conda' plugin."
    )
    assert (
        exception.get_resolution()
        == "Ensure running the build on a supported architecture for this plugin."
    )
    assert exception.get_details() is None
    assert exception.get_exit_code() == 2
    assert exception.get_docs_url() is None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import http.server
import logging
import os
import stat
import threading
from unittest import mock

import apt
import fixtures
import progressbar
import testscenarios
import testtools

from snapcraft.internal import common, steps
from tests import fake_servers, fixture_setup
from tests.file_utils import get_snapcraft_path
from tests.unit.part_loader import load_part


class ContainsList(list):
    def __eq__(self, other):
        return all([i[0] in i[1] for i in zip(self, other)])


class MockOptions:
    def __init__(
        self,
        source=None,
        source_type=None,
        source_branch=None,
        source_tag=None,
        source_subdir=None,
        source_depth=None,
        source_commit=None,
        source_checksum=None,
        disable_parallel=False,
    ):
        self.source = source
        self.source_type = source_type
        self.source_depth = source_depth
        self.source_branch = source_branch
        self.source_commit = source_commit
        self.source_tag = source_tag
        self.source_subdir = source_subdir
        self.disable_parallel = disable_parallel


class IsExecutable:
    """Match if a file path is executable."""

    def __str__(self):
        return "IsExecutable()"

    def match(self, file_path):
        if not os.stat(file_path).st_mode & stat.S_IEXEC:
            return testtools.matchers.Mismatch(
                "Expected {!r} to be executable, but it was not".format(file_path)
            )
        return None


class LinkExists:
    """Match if a file path is a symlink."""

    def __init__(self, expected_target=None):
        self._expected_target = expected_target

    def __str__(self):
        return "LinkExists()"

    def match(self, file_path):
        if not os.path.exists(file_path):
            return testtools.matchers.Mismatch(
                "Expected {!r} to be a symlink, but it doesn't exist".format(file_path)
            )

        if not os.path.islink(file_path):
            return testtools.matchers.Mismatch(
                "Expected {!r} to be a symlink, but it was not".format(file_path)
            )

        target = os.readlink(file_path)
        if target != self._expected_target:
            return testtools.matchers.Mismatch(
                "Expected {!r} to be a symlink pointing to {!r}, but it was "
                "pointing to {!r}".format(file_path, self._expected_target, target)
            )

        return None


class TestCase(testscenarios.WithScenarios, testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.fake_snapd = fixture_setup.FakeSnapd()
        cls.fake_snapd.setUp()

    @classmethod
    def tearDownClass(cls):
        cls.fake_snapd.cleanUp()

    def setUp(self):
        super().setUp()
        temp_cwd_fixture = fixture_setup.TempCWD()
        self.useFixture(temp_cwd_fixture)
        self.path = temp_cwd_fixture.path

        # Use a separate path for XDG dirs, or changes there may be detected as
        # source changes.
        self.xdg_path = self.useFixture(fixtures.TempDir()).path
        self.useFixture(fixture_setup.TempXDG(self.xdg_path))
        self.fake_terminal = fixture_setup.FakeTerminal()
        self.useFixture(self.fake_terminal)
        # Some tests will directly or indirectly change the plugindir, which
        # is a module variable. Make sure that it is returned to the original
        # value when a test ends.
        self.addCleanup(common.set_plugindir, common.get_plugindir())
        self.addCleanup(common.set_schemadir, common.get_schemadir())
        self.addCleanup(common.set_extensionsdir, common.get_extensionsdir())
        self.addCleanup(common.set_keyringsdir, common.get_keyringsdir())
        self.addCleanup(common.reset_env)
        common.set_schemadir(os.path.join(get_snapcraft_path(), "schema"))
        self.fake_logger = fixtures.FakeLogger(level=logging.ERROR)
        self.useFixture(self.fake_logger)

        # Some tests will change the apt Dir::Etc::Trusted and
        # Dir::Etc::TrustedParts directories. Make sure they're properly reset.
        self.addCleanup(
            apt.apt_pkg.config.set,
            "Dir::Etc::Trusted",
            apt.apt_pkg.config.find_file("Dir::Etc::Trusted"),
        )
        self.addCleanup(
            apt.apt_pkg.config.set,
            "Dir::Etc::TrustedParts",
            apt.apt_pkg.config.find_file("Dir::Etc::TrustedParts"),
        )

        patcher = mock.patch("os.sched_getaffinity")
        self.cpu_count = patcher.start()
        self.cpu_count.return_value = {1, 2}
        self.addCleanup(patcher.stop)

        # We do not want the paths to affect every test we have.
        patcher = mock.patch(
            "snapcraft.file_utils.get_snap_tool_path", side_effect=lambda x: x
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch(
            "snapcraft.internal.indicators.ProgressBar", new=SilentProgressBar
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        # These are what we expect by default
        self.snap_dir = os.path.join(os.getcwd(), "snap")
        self.prime_dir = os.path.join(os.getcwd(), "prime")
        self.stage_dir = os.path.join(os.getcwd(), "stage")
        self.parts_dir = os.path.join(os.getcwd(), "parts")
        self.local_plugins_dir = os.path.join(self.snap_dir, "plugins")

        # Use this host to run through the lifecycle tests
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )

        # Make sure snap installation does the right thing
        self.fake_snapd.installed_snaps = [
            dict(name="core16", channel="stable", revision="10"),
            dict(name="core18", channel="stable", revision="10"),
        ]
        self.fake_snapd.snaps_result = [
            dict(name="core16", channel="stable", revision="10"),
            dict(name="core18", channel="stable", revision="10"),
        ]
        self.fake_snapd.find_result = [
            dict(
                core16=dict(
                    channel="stable",
                    channels={"latest/stable": dict(confinement="strict")},
                )
            ),
            dict(
                core18=dict(
                    channel="stable",
                    channels={"latest/stable": dict(confinement="strict")},
                )
            ),
        ]
        self.fake_snapd.snap_details_func = None

        self.fake_snap_command = fixture_setup.FakeSnapCommand()
        self.useFixture(self.fake_snap_command)

        # Avoid installing patchelf in the tests
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_NO_PATCHELF", "1"))

        # Disable Sentry reporting for tests, otherwise they'll hang waiting
        # for input
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_ERROR_REPORTING", "false")
        )

        # Don't let the managed host variable leak into tests
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_MANAGED_HOST"))

        machine = os.environ.get("SNAPCRAFT_TEST_MOCK_MACHINE", None)
        self.base_environment = fixture_setup.FakeBaseEnvironment(machine=machine)
        self.useFixture(self.base_environment)

        # Make sure "SNAPCRAFT_ENABLE_DEVELOPER_DEBUG" is reset between tests
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_DEVELOPER_DEBUG")
        )
        self.useFixture(fixture_setup.FakeSnapcraftctl())

        # Don't let host SNAPCRAFT_BUILD_INFO variable leak into tests
        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_INFO"))

    def make_snapcraft_yaml(self, content, encoding="utf-8", location=""):
        snap_dir = os.path.join(location, "snap")
        os.makedirs(snap_dir, exist_ok=True)
        snapcraft_yaml = os.path.join(snap_dir, "snapcraft.yaml")
        with open(snapcraft_yaml, "w", encoding=encoding) as fp:
            fp.write(content)
        return snapcraft_yaml

    def verify_state(self, part_name, state_dir, expected_step_name):
        self.assertTrue(
            os.path.isdir(state_dir),
            "Expected state directory for {}".format(part_name),
        )

        # Expect every step up to and including the specified one to be run
        step = steps.get_step_by_name(expected_step_name)
        for step in step.previous_steps() + [step]:
            self.assertTrue(
                os.path.exists(os.path.join(state_dir, step.name)),
                "Expected {!r} to be run for {}".format(step.name, part_name),
            )

    def load_part(
        self,
        part_name,
        plugin_name=None,
        part_properties=None,
        project=None,
        stage_packages_repo=None,
        snap_name="test-snap",
        base="core18",
        build_base=None,
        confinement="strict",
        snap_type="app",
    ):
        return load_part(
            part_name=part_name,
            plugin_name=plugin_name,
            part_properties=part_properties,
            project=project,
            stage_packages_repo=stage_packages_repo,
            snap_name=snap_name,
            base=base,
            build_base=build_base,
            confinement=confinement,
            snap_type=snap_type,
        )


class TestWithFakeRemoteParts(TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(fixture_setup.FakeParts())


class FakeFileHTTPServerBasedTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))
        self.server = http.server.HTTPServer(
            ("127.0.0.1", 0), fake_servers.FakeFileHTTPRequestHandler
        )
        server_thread = threading.Thread(target=self.server.serve_forever)
        self.addCleanup(server_thread.join)
        self.addCleanup(self.server.server_close)
        self.addCleanup(self.server.shutdown)
        server_thread.start()


class SilentProgressBar(progressbar.ProgressBar):
    """A progress bar causing no spurious output during tests."""

    def start(self):
        pass

    def update(self, value=None):
        pass

    def finish(self):
        pass



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import subprocess
from unittest import mock
from unittest.mock import call

import gnupg
import pytest

from snapcraft.internal.meta.package_repository import (
    PackageRepositoryApt,
    PackageRepositoryAptPpa,
)
from snapcraft.internal.repo import apt_ppa, errors
from snapcraft.internal.repo.apt_key_manager import AptKeyManager


@pytest.fixture(autouse=True)
def mock_environ_copy():
    with mock.patch("os.environ.copy") as m:
        yield m


@pytest.fixture(autouse=True)
def mock_gnupg(tmp_path, autouse=True):
    with mock.patch("gnupg.GPG", spec=gnupg.GPG) as m:
        m.return_value.import_keys.return_value.fingerprints = [
            "FAKE-KEY-ID-FROM-GNUPG"
        ]
        yield m


@pytest.fixture(autouse=True)
def mock_run():
    with mock.patch("subprocess.run", spec=subprocess.run) as m:
        yield m


@pytest.fixture(autouse=True)
def mock_apt_ppa_get_signing_key():
    with mock.patch(
        "snapcraft.internal.repo.apt_ppa.get_launchpad_ppa_key_id",
        spec=apt_ppa.get_launchpad_ppa_key_id,
        return_value="FAKE-PPA-SIGNING-KEY",
    ) as m:
        yield m


@pytest.fixture
def key_assets(tmp_path):
    key_assets = tmp_path / "key-assets"
    key_assets.mkdir(parents=True)
    yield key_assets


@pytest.fixture
def gpg_keyring(tmp_path):
    yield tmp_path / "keyring.gpg"


@pytest.fixture
def apt_gpg(key_assets, gpg_keyring):
    yield AptKeyManager(
        gpg_keyring=gpg_keyring, key_assets=key_assets,
    )


def test_find_asset(
    apt_gpg, key_assets,
):
    key_id = "8" * 40
    expected_key_path = key_assets / ("8" * 8 + ".asc")
    expected_key_path.write_text("key")

    key_path = apt_gpg.find_asset_with_key_id(key_id=key_id)

    assert key_path == expected_key_path


def test_find_asset_none(apt_gpg,):
    key_path = apt_gpg.find_asset_with_key_id(key_id="foo")

    assert key_path is None


def test_get_key_fingerprints(
    apt_gpg, mock_gnupg,
):
    with mock.patch("tempfile.NamedTemporaryFile") as m:
        m.return_value.__enter__.return_value.name = "/tmp/foo"
        ids = apt_gpg.get_key_fingerprints(key="8" * 40)

    assert ids == ["FAKE-KEY-ID-FROM-GNUPG"]
    assert mock_gnupg.mock_calls == [
        call(keyring="/tmp/foo"),
        call().import_keys(key_data="8888888888888888888888888888888888888888"),
    ]


@pytest.mark.parametrize(
    "stdout,expected",
    [
        (b"nothing exported", False),
        (b"BEGIN PGP PUBLIC KEY BLOCK", True),
        (b"invalid", False),
    ],
)
def test_is_key_installed(
    stdout, expected, apt_gpg, mock_run,
):
    mock_run.return_value.stdout = stdout

    is_installed = apt_gpg.is_key_installed(key_id="foo")

    assert is_installed is expected
    assert mock_run.mock_calls == [
        call(
            ["sudo", "apt-key", "export", "foo"],
            check=True,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
    ]


def test_is_key_installed_with_apt_key_failure(
    apt_gpg, mock_run,
):
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["apt-key"], returncode=1, output=b"some error"
    )

    is_installed = apt_gpg.is_key_installed(key_id="foo")

    assert is_installed is False


def test_install_key(
    apt_gpg, gpg_keyring, mock_run,
):
    key = "some-fake-key"
    apt_gpg.install_key(key=key)

    assert mock_run.mock_calls == [
        call(
            ["sudo", "apt-key", "--keyring", str(gpg_keyring), "add", "-"],
            check=True,
            env={"LANG": "C.UTF-8"},
            input=b"some-fake-key",
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
    ]


def test_install_key_with_apt_key_failure(apt_gpg, mock_run):
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["foo"], returncode=1, output=b"some error"
    )

    with pytest.raises(errors.AptGPGKeyInstallError) as exc_info:
        apt_gpg.install_key(key="FAKEKEY")

    assert exc_info.value._output == "some error"
    assert exc_info.value._key == "FAKEKEY"


def test_install_key_from_keyserver(apt_gpg, gpg_keyring, mock_run):
    apt_gpg.install_key_from_keyserver(key_id="FAKE_KEYID", key_server="key.server")

    assert mock_run.mock_calls == [
        call(
            [
                "sudo",
                "apt-key",
                "--keyring",
                str(gpg_keyring),
                "adv",
                "--keyserver",
                "key.server",
                "--recv-keys",
                "FAKE_KEYID",
            ],
            check=True,
            env={"LANG": "C.UTF-8"},
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
    ]


def test_install_key_from_keyserver_with_apt_key_failure(
    apt_gpg, gpg_keyring, mock_run
):
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["apt-key"], returncode=1, output=b"some error"
    )

    with pytest.raises(errors.AptGPGKeyInstallError) as exc_info:
        apt_gpg.install_key_from_keyserver(
            key_id="fake-key-id", key_server="fake-server"
        )

    assert exc_info.value._output == "some error"
    assert exc_info.value._key_id == "fake-key-id"


@mock.patch("snapcraft.internal.repo.apt_key_manager.AptKeyManager.is_key_installed")
@pytest.mark.parametrize(
    "is_installed", [True, False],
)
def test_install_package_repository_key_already_installed(
    mock_is_key_installed, is_installed, apt_gpg,
):
    mock_is_key_installed.return_value = is_installed
    package_repo = PackageRepositoryApt(
        components=["main", "multiverse"],
        key_id="8" * 40,
        key_server="xkeyserver.com",
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is not is_installed


@mock.patch(
    "snapcraft.internal.repo.apt_key_manager.AptKeyManager.is_key_installed",
    return_value=False,
)
@mock.patch("snapcraft.internal.repo.apt_key_manager.AptKeyManager.install_key")
def test_install_package_repository_key_from_asset(
    mock_install_key, mock_is_key_installed, apt_gpg, key_assets,
):
    key_id = "123456789012345678901234567890123456AABB"
    expected_key_path = key_assets / "3456AABB.asc"
    expected_key_path.write_text("key-data")

    package_repo = PackageRepositoryApt(
        components=["main", "multiverse"],
        key_id=key_id,
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key.mock_calls == [call(key="key-data")]


@mock.patch(
    "snapcraft.internal.repo.apt_key_manager.AptKeyManager.is_key_installed",
    return_value=False,
)
@mock.patch(
    "snapcraft.internal.repo.apt_key_manager.AptKeyManager.install_key_from_keyserver"
)
def test_install_package_repository_key_apt_from_keyserver(
    mock_install_key_from_keyserver, mock_is_key_installed, apt_gpg,
):
    key_id = "8" * 40

    package_repo = PackageRepositoryApt(
        components=["main", "multiverse"],
        key_id=key_id,
        key_server="key.server",
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key_from_keyserver.mock_calls == [
        call(key_id=key_id, key_server="key.server")
    ]


@mock.patch(
    "snapcraft.internal.repo.apt_key_manager.AptKeyManager.is_key_installed",
    return_value=False,
)
@mock.patch(
    "snapcraft.internal.repo.apt_key_manager.AptKeyManager.install_key_from_keyserver"
)
def test_install_package_repository_key_ppa_from_keyserver(
    mock_install_key_from_keyserver, mock_is_key_installed, apt_gpg,
):
    package_repo = PackageRepositoryAptPpa(ppa="test/ppa",)

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key_from_keyserver.mock_calls == [
        call(key_id="FAKE-PPA-SIGNING-KEY", key_server="keyserver.ubuntu.com")
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.plugins.v2.cmake import CMakePlugin


def test_schema():
    assert CMakePlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "cmake-parameters": {
                "type": "array",
                "uniqueItems": True,
                "items": {"type": "string"},
                "default": [],
            },
            "cmake-generator": {
                "type": "string",
                "enum": ["Unix Makefiles", "Ninja"],
                "default": "Unix Makefiles",
            },
        },
    }


def test_get_build_packages():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Unix Makefiles"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_packages() == {"gcc", "cmake"}


def test_get_build_packages_ninja():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Ninja"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_packages() == {"gcc", "cmake", "ninja-build"}


def test_get_build_environment():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Unix Makefiles"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_environment() == {
        "CMAKE_PREFIX_PATH": "${SNAPCRAFT_STAGE}",
        "SNAPCRAFT_CMAKE_ARGS": "",
    }


def test_get_build_commands():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Unix Makefiles"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'cmake "${SNAPCRAFT_PART_SRC_WORK}" -G "Unix Makefiles" ${SNAPCRAFT_CMAKE_ARGS}',
        'cmake --build . -- -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'DESTDIR="${SNAPCRAFT_PART_INSTALL}" cmake --build . --target install',
    ]


def test_get_build_commands_ninja():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Ninja"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'cmake "${SNAPCRAFT_PART_SRC_WORK}" -G "Ninja" ${SNAPCRAFT_CMAKE_ARGS}',
        'cmake --build . -- -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'DESTDIR="${SNAPCRAFT_PART_INSTALL}" cmake --build . --target install',
    ]


def test_get_build_commands_with_cmake_parameters():
    class Options:
        cmake_parameters = [
            "-DVERBOSE=1",
            "-DCMAKE_INSTALL_PREFIX=/foo",
            '-DCMAKE_SPACED_ARGS="foo bar"',
            '-DCMAKE_USING_ENV="$SNAPCRAFT_PART_INSTALL"/bar',
        ]
        cmake_generator = "Unix Makefiles"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'cmake "${SNAPCRAFT_PART_SRC_WORK}" -G "Unix Makefiles" '
        "${SNAPCRAFT_CMAKE_ARGS} "
        "-DVERBOSE=1 "
        "-DCMAKE_INSTALL_PREFIX=/foo "
        '-DCMAKE_SPACED_ARGS="foo bar" '
        '-DCMAKE_USING_ENV="$SNAPCRAFT_PART_INSTALL"/bar',
        'cmake --build . -- -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'DESTDIR="${SNAPCRAFT_PART_INSTALL}" cmake --build . --target install',
    ]


def test_out_of_source_build():
    class Options:
        cmake_parameters = list()
        cmake_generator = "Unix Makefiles"

    plugin = CMakePlugin(part_name="my-part", options=Options())

    assert plugin.out_of_source_build is True



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import os
import shutil
from unittest import mock

from testtools.matchers import DirExists, Equals, FileContains, FileExists, Not

from snapcraft.internal import common, errors, sources
from tests import unit


class TestLocal(unit.TestCase):
    @mock.patch("snapcraft.internal.sources._local.glob.glob")
    def test_pull_does_not_change_snapcraft_files_list(self, mock_glob):
        # Regression test for https://bugs.launchpad.net/snapcraft/+bug/1614913
        # Verify that SNAPCRAFT_FILES was not modified by the pull when there
        # are files to ignore.
        snapcraft_files_before_pull = copy.copy(common.SNAPCRAFT_FILES)
        mock_glob.return_value = ["a.snap", "b.snap", "c.snap"]

        local = sources.Local(".", "destination")
        local.pull()

        self.assertThat(snapcraft_files_before_pull, Equals(common.SNAPCRAFT_FILES))

    def test_pull_with_existing_empty_source_dir_creates_hardlinks(self):
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        os.mkdir("destination")

        local = sources.Local("src", "destination")
        local.pull()

        # Verify that the directories are not symlinks, but the file is a
        # hardlink.
        self.assertFalse(os.path.islink("destination"))
        self.assertFalse(os.path.islink(os.path.join("destination", "dir")))
        self.assertGreater(
            os.stat(os.path.join("destination", "dir", "file")).st_nlink, 1
        )

    def test_pull_with_existing_source_tree_creates_hardlinks(self):
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        os.mkdir("destination")
        open(os.path.join("destination", "existing-file"), "w").close()

        local = sources.Local("src", "destination")
        local.pull()

        # Verify that the directories are not symlinks, but the file is a
        # hardlink. Also verify that existing-file still exists.
        self.assertFalse(os.path.islink("destination"))
        self.assertFalse(os.path.islink(os.path.join("destination", "dir")))
        self.assertThat(os.path.join("destination", "existing-file"), FileExists())
        self.assertGreater(
            os.stat(os.path.join("destination", "dir", "file")).st_nlink, 1
        )

    def test_pull_with_existing_source_link_error(self):
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        # Note that this is a symlink now instead of a directory
        os.symlink("dummy", "destination")

        local = sources.Local("src", "destination")
        self.assertRaises(errors.SnapcraftEnvironmentError, local.pull)

    def test_pull_with_existing_source_file_error(self):
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        # Note that this is a file now instead of a directory
        open("destination", "w").close()

        local = sources.Local("src", "destination")
        self.assertRaises(errors.SnapcraftEnvironmentError, local.pull)

    def test_pulling_twice_with_existing_source_dir_recreates_hardlinks(self):
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        os.mkdir("destination")

        local = sources.Local("src", "destination")
        local.pull()
        local.pull()

        # Verify that the directories are not symlinks, but the file is a
        # hardlink.
        self.assertFalse(os.path.islink("destination"))
        self.assertFalse(os.path.islink(os.path.join("destination", "dir")))
        self.assertGreater(
            os.stat(os.path.join("destination", "dir", "file")).st_nlink, 1
        )

    def test_pull_ignores_snapcraft_specific_data(self):
        # Make the snapcraft-specific directories
        os.makedirs(os.path.join("src", "parts"))
        os.makedirs(os.path.join("src", "stage"))
        os.makedirs(os.path.join("src", "prime"))
        os.makedirs(os.path.join("src", ".snapcraft"))
        os.makedirs(os.path.join("src", "snap"))

        # Make the snapcraft.yaml (and hidden one) and a built snap
        open(os.path.join("src", "snapcraft.yaml"), "w").close()
        open(os.path.join("src", ".snapcraft.yaml"), "w").close()
        open(os.path.join("src", "foo.snap"), "w").close()

        # Make the global state cache
        open(os.path.join("src", ".snapcraft", "state"), "w").close()

        # Now make some real files
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()

        os.mkdir("destination")

        local = sources.Local("src", "destination")
        local.pull()

        # Verify that the snapcraft-specific stuff got filtered out
        self.assertThat(os.path.join("destination", "parts"), Not(DirExists()))
        self.assertThat(os.path.join("destination", "stage"), Not(DirExists()))
        self.assertThat(os.path.join("destination", "prime"), Not(DirExists()))

        self.assertThat(os.path.join("destination", "snap"), DirExists())
        self.assertThat(os.path.join("destination", ".snapcraft.yaml"), FileExists())
        self.assertThat(os.path.join("destination", "snapcraft.yaml"), FileExists())

        self.assertThat(os.path.join("destination", "foo.snap"), Not(FileExists()))

        # Verify that the real stuff made it in.
        self.assertFalse(os.path.islink("destination"))
        self.assertFalse(os.path.islink(os.path.join("destination", "dir")))
        self.assertGreater(
            os.stat(os.path.join("destination", "dir", "file")).st_nlink, 1
        )

    def test_pull_keeps_symlinks(self):
        # Create a source containing a directory, a file and symlinks to both.
        os.makedirs(os.path.join("src", "dir"))
        open(os.path.join("src", "dir", "file"), "w").close()
        os.symlink("dir", os.path.join("src", "dir_symlink"))
        os.symlink("file", os.path.join("src", "dir", "file_symlink"))

        local = sources.Local("src", "destination")
        local.pull()

        # Verify that both the file and the directory symlinks were kept.
        self.expectThat(os.path.join("destination", "dir"), DirExists())
        self.expectThat(
            os.path.join("destination", "dir_symlink"), unit.LinkExists("dir")
        )
        self.expectThat(os.path.join("destination", "dir", "file"), FileExists())
        self.expectThat(
            os.path.join("destination", "dir", "file_symlink"), unit.LinkExists("file")
        )

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["local"] is sources.Local)


class TestLocalIgnores:
    """Verify that the snapcraft root dir does not get copied into itself."""

    scenarios = [
        (f, dict(snapcraft_dir=f)) for f in common.SNAPCRAFT_FILES if "." not in f
    ]

    def test_pull_with_source_the_parent_of_current_dir(
        self, tmp_work_path, snapcraft_dir
    ):
        os.makedirs("subdir")

        cwd = tmp_work_path.as_posix()
        os.chdir("subdir")
        source_dir = os.path.join(snapcraft_dir, "foo_src")
        local = sources.Local("..", source_dir)
        local.pull()
        os.chdir(cwd)

        assert not os.path.exists(
            os.path.join("subdir", source_dir, "subdir", snapcraft_dir)
        )
        assert "subdir" in os.listdir(os.path.join("subdir", source_dir))

    def test_pull_with_source_a_parent_of_current_dir(
        self, tmp_work_path, snapcraft_dir
    ):
        subdir = os.path.join("subdir", "subsubdir", "subsubsubdir")
        os.makedirs(subdir)

        cwd = tmp_work_path.as_posix()
        os.chdir(subdir)
        source = "../" * (subdir.count(os.sep) + 1)
        source_dir = os.path.join(snapcraft_dir, "foo_src")
        local = sources.Local(source, source_dir)
        local.pull()
        os.chdir(cwd)
        assert not os.path.exists(
            os.path.join(subdir, source_dir, subdir, snapcraft_dir)
        )
        assert os.path.basename(subdir) in os.listdir(
            os.path.join(subdir, source_dir, os.path.dirname(subdir))
        )


class TestLocalUpdate(unit.TestCase):
    """Verify that the local source can detect changes and update."""

    def test_file_modified(self):
        source = "source"
        destination = "destination"
        os.mkdir(source)
        os.mkdir(destination)

        with open(os.path.join(source, "file"), "w") as f:
            f.write("1")

        # Now make a reference file with a timestamp later than the file was
        # created. We'll ensure this by setting it ourselves
        shutil.copy2(os.path.join(source, "file"), "reference")
        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime("reference", (access_time, modify_time + 1))

        local = sources.Local(source, destination)
        local.pull()
        self.assertFalse(
            local.check("reference"), "Expected no updates to be available"
        )
        self.assertThat(os.path.join(destination, "file"), FileContains("1"))

        # Now update the file in source, and make sure it has a timestamp
        # later than our reference (this whole test happens too fast)
        with open(os.path.join(source, "file"), "w") as f:
            f.write("2")

        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime(os.path.join(source, "file"), (access_time, modify_time + 1))

        self.assertTrue(local.check("reference"), "Expected update to be available")

        local.update()
        self.assertThat(os.path.join(destination, "file"), FileContains("2"))

    def test_file_added(self):
        source = "source"
        destination = "destination"
        os.mkdir(source)
        os.mkdir(destination)

        with open(os.path.join(source, "file1"), "w") as f:
            f.write("1")

        # Now make a reference file with a timestamp later than the file was
        # created. We'll ensure this by setting it ourselves
        shutil.copy2(os.path.join(source, "file1"), "reference")
        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime("reference", (access_time, modify_time + 1))

        local = sources.Local(source, destination)
        local.pull()
        self.assertFalse(
            local.check("reference"), "Expected no updates to be available"
        )
        self.assertThat(os.path.join(destination, "file1"), FileExists())

        # Now add a new file, and make sure it has a timestamp
        # later than our reference (this whole test happens too fast)
        with open(os.path.join(source, "file2"), "w") as f:
            f.write("2")

        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime(os.path.join(source, "file2"), (access_time, modify_time + 1))

        self.assertTrue(local.check("reference"), "Expected update to be available")

        local.update()
        self.assertThat(os.path.join(destination, "file2"), FileExists())

    def test_directory_modified(self):
        source = "source"
        source_dir = os.path.join(source, "dir")
        destination = "destination"
        os.makedirs(source_dir)
        os.mkdir(destination)

        with open(os.path.join(source_dir, "file1"), "w") as f:
            f.write("1")

        # Now make a reference file with a timestamp later than the file was
        # created. We'll ensure this by setting it ourselves
        shutil.copy2(os.path.join(source_dir, "file1"), "reference")
        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime("reference", (access_time, modify_time + 1))

        local = sources.Local(source, destination)
        local.pull()
        self.assertFalse(
            local.check("reference"), "Expected no updates to be available"
        )
        self.assertThat(os.path.join(destination, "dir", "file1"), FileExists())

        # Now add a new file to the directory, and make sure it has a timestamp
        # later than our reference (this whole test happens too fast)
        with open(os.path.join(source_dir, "file2"), "w") as f:
            f.write("2")

        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime(os.path.join(source_dir, "file2"), (access_time, modify_time + 1))

        self.assertTrue(local.check("reference"), "Expected update to be available")

        local.update()
        self.assertThat(os.path.join(destination, "dir", "file2"), FileExists())


class TestLocalUpdateSnapcraftYaml:

    scenarios = [
        ("snapcraft.yaml", dict(snapcraft_file="snapcraft.yaml")),
        (".snapcraft.yaml", dict(snapcraft_file=".snapcraft.yaml")),
        (
            "snap/snapcraft.yaml",
            dict(snapcraft_file=os.path.join("snap", "snapcraft.yaml")),
        ),
        ("snap/<content>", dict(snapcraft_file=os.path.join("snap", "test-file"))),
    ]

    def test_snapcraft_yaml_modification_ignored(self, tmp_work_path, snapcraft_file):
        source = "source"
        destination = "destination"
        snapcraft_source_path = os.path.join(source, snapcraft_file)
        snapcraft_destination_path = os.path.join(destination, snapcraft_file)
        os.makedirs(os.path.dirname(snapcraft_source_path))
        os.mkdir(destination)

        with open(snapcraft_source_path, "w") as f:
            f.write("1")

        # Now make a reference file with a timestamp later than the file was
        # created. We'll ensure this by setting it ourselves
        shutil.copy2(snapcraft_source_path, "reference")
        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime("reference", (access_time, modify_time + 1))

        local = sources.Local(source, destination)
        local.pull()
        assert not local.check("reference")
        assert os.path.exists(snapcraft_destination_path)

        # Now add a new file to the directory, and make sure it has a timestamp
        # later than our reference (this whole test happens too fast)
        with open(snapcraft_source_path, "w") as f:
            f.write("2")

        access_time = os.stat("reference").st_atime
        modify_time = os.stat("reference").st_mtime
        os.utime(snapcraft_source_path, (access_time, modify_time + 1))

        assert not local.check("reference")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import pytest


@pytest.fixture
def mock_echo_error():
    """Return a mock for snapcraft.cli.echo.error."""
    patcher = mock.patch("snapcraft.cli.echo.error")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_sys_exit():
    """Return a mock for sys.exit."""
    patcher = mock.patch("sys.exit")
    yield patcher.start()
    patcher.stop()
